#! COCO image captioning dataset

import os

from clarifai.datasets.upload.base import ClarifaiDataLoader

from ..features import VisualClassificationFeatures

#pycocotools is a dependency for this loader
try:
  from pycocotools.coco import COCO
except ImportError:
  raise ImportError("Could not import pycocotools package. "
                    "Please do `pip install 'clarifai[all]'` to import pycocotools.")


class COCOCaptionsDataLoader(ClarifaiDataLoader):
  """COCO Image Captioning Dataset."""

  def __init__(self, images_dir, label_filepath):
    """
    Args:
      images_dir: Directory containing the images.
      label_filepath: Path to the COCO annotation file.
    """
    self.images_dir = images_dir
    self.label_filepath = label_filepath

    self.map_ids = {}
    self.load_data()

  @property
  def task(self):
    return "visual_captioning"

  def load_data(self) -> None:
    self.coco = COCO(self.label_filepath)
    self.map_ids = {i: img_id for i, img_id in enumerate(list(self.coco.imgs.keys()))}

  def __len__(self):
    return len(self.coco.imgs)

  def __getitem__(self, index):
    value = self.coco.imgs[self.map_ids[index]]
    image_path = os.path.join(self.images_dir, value['file_name'])
    annots = []

    input_ann_ids = self.coco.getAnnIds(imgIds=[value['id']])
    input_anns = self.coco.loadAnns(input_ann_ids)

    for ann in input_anns:
      annots.append(ann['caption'])

    return VisualClassificationFeatures(image_path, labels=annots[0], id=str(value['id']))



import importlib
import os
import pkgutil

import click
import yaml


def from_yaml(filename: str):
  try:
    with open(filename, 'r') as f:
      return yaml.safe_load(f)
  except yaml.YAMLError as e:
    click.echo(f"Error reading YAML file: {e}", err=True)
    return {}


def dump_yaml(data, filename: str):
  try:
    with open(filename, 'w') as f:
      yaml.dump(data, f)
  except Exception as e:
    click.echo(f"Error writing YAML file: {e}", err=True)


def set_base_url(env):
  environments = {
      'prod': 'https://api.clarifai.com',
      'staging': 'https://api-staging.clarifai.com',
      'dev': 'https://api-dev.clarifai.com'
  }

  if env in environments:
    return environments[env]
  else:
    raise ValueError("Invalid environment. Please choose from 'prod', 'staging', 'dev'.")


# Dynamically find and import all command modules from the cli directory
def load_command_modules():
  package_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'cli')

  for _, module_name, _ in pkgutil.iter_modules([package_dir]):
    if module_name != 'base':  # Skip the base.py file itself
      importlib.import_module(f'clarifai.cli.{module_name}')



import os
import uuid
from typing import Any, Dict, List

from clarifai.errors import UserError


class Chunker:
  """Split an input sequence into small chunks."""

  def __init__(self, seq: List, size: int) -> None:
    self.seq = seq
    self.size = size

  def chunk(self) -> List[List]:
    """Chunk input sequence."""
    return [self.seq[pos:pos + self.size] for pos in range(0, len(self.seq), self.size)]


class BackoffIterator:
  """Iterator that returns a sequence of backoff values."""

  def __init__(self, count=0):
    self.count = count

  def __iter__(self):
    return self

  def __next__(self):
    self.count += 1
    return 0.1 * (1.3**self.count)


def get_from_dict_or_env(key: str, env_key: str, **data) -> str:
  """Get a value from a dictionary or an environment variable."""
  if key in data and data[key]:
    return data[key]
  else:
    return get_from_env(key, env_key)


def get_from_env(key: str, env_key: str) -> str:
  """Get a value from a dictionary or an environment variable."""
  if env_key in os.environ and os.environ[env_key]:
    return os.environ[env_key]
  else:
    raise UserError(f"Did not find `{key}`, please add an environment variable"
                    f" `{env_key}` which contains it, or pass"
                    f"  `{key}` as a named parameter.")


def concept_relations_accumulation(relations_dict: Dict[str, Any], subject_concept: str,
                                   object_concept: str, predicate: str) -> Dict[str, Any]:
  """Append the concept relation to relations dict based on its predicate.

    Args:
        relations_dict (dict): A dict of concept relations info.
    """
  if predicate == 'hyponym':
    if object_concept in relations_dict:
      relations_dict[object_concept].append(subject_concept)
    else:
      relations_dict[object_concept] = [subject_concept]
  elif predicate == 'hypernym':
    if subject_concept in relations_dict:
      relations_dict[subject_concept].append(object_concept)
    else:
      relations_dict[subject_concept] = [object_concept]
  else:
    relations_dict[object_concept] = []
    relations_dict[subject_concept] = []
  return relations_dict


def get_uuid(val: int) -> str:
  """Generates a UUID."""
  return uuid.uuid4().hex[:val]



import os

import click

from ..utils.cli import dump_yaml, from_yaml, load_command_modules, set_base_url


@click.group()
@click.pass_context
def cli(ctx):
  """Clarifai CLI"""
  ctx.ensure_object(dict)
  config_path = 'config.yaml'
  if os.path.exists(config_path):
    ctx.obj = from_yaml(config_path)
  else:
    ctx.obj = {}


@cli.command()
@click.option('--config', type=click.Path(), required=False, help='Path to the config file')
@click.option(
    '-e',
    '--env',
    required=False,
    help='Environment to use, choose from prod, staging and dev',
    type=click.Choice(['prod', 'staging', 'dev']))
@click.option('--user_id', required=False, help='User ID')
@click.pass_context
def login(ctx, config, env, user_id):
  """Login command to set PAT and other configurations."""

  if config and os.path.exists(config):
    ctx.obj = from_yaml(config)

  if 'pat' in ctx.obj:
    os.environ["CLARIFAI_PAT"] = ctx.obj['pat']
    click.echo("Loaded PAT from config file.")
  elif 'CLARIFAI_PAT' in os.environ:
    ctx.obj['pat'] = os.environ["CLARIFAI_PAT"]
    click.echo("Loaded PAT from environment variable.")
  else:
    _pat = click.prompt(
        "Get your PAT from https://clarifai.com/settings/security and pass it here", type=str)
    os.environ["CLARIFAI_PAT"] = _pat
    ctx.obj['pat'] = _pat
    click.echo("PAT saved successfully.")

  if user_id:
    ctx.obj['user_id'] = user_id
    os.environ["CLARIFAI_USER_ID"] = ctx.obj['user_id']
  elif 'user_id' in ctx.obj or 'CLARIFAI_USER_ID' in os.environ:
    ctx.obj['user_id'] = ctx.obj.get('user_id', os.environ["CLARIFAI_USER_ID"])
    os.environ["CLARIFAI_USER_ID"] = ctx.obj['user_id']

  if env:
    ctx.obj['env'] = env
    ctx.obj['base_url'] = set_base_url(env)
    os.environ["CLARIFAI_API_BASE"] = ctx.obj['base_url']
  elif 'env' in ctx.obj:
    ctx.obj['env'] = ctx.obj.get('env', "prod")
    ctx.obj['base_url'] = set_base_url(ctx.obj['env'])
    os.environ["CLARIFAI_API_BASE"] = ctx.obj['base_url']
  elif 'CLARIFAI_API_BASE' in os.environ:
    ctx.obj['base_url'] = os.environ["CLARIFAI_API_BASE"]

  dump_yaml(ctx.obj, 'config.yaml')


# Import the CLI commands to register them
load_command_modules()

if __name__ == '__main__':
  cli()



import concurrent.futures

import fsspec

from clarifai.utils.logging import logger


def download_input(input):
  """
  This function will download any urls that are not already bytes.
  """
  if input.data.image.url and not input.data.image.base64:
    # Download the image
    with fsspec.open(input.data.image.url, 'rb') as f:
      input.data.image.base64 = f.read()
  if input.data.video.url and not input.data.video.base64:
    # Download the video
    with fsspec.open(input.data.video.url, 'rb') as f:
      input.data.video.base64 = f.read()
  if input.data.audio.url and not input.data.audio.base64:
    # Download the audio
    with fsspec.open(input.data.audio.url, 'rb') as f:
      input.data.audio.base64 = f.read()
  if input.data.text.url and not input.data.text.raw:
    # Download the text
    with fsspec.open(input.data.text.url, 'r') as f:
      input.data.text.raw = f.read()


def ensure_urls_downloaded(request, max_threads=128):
  """
  This function will download any urls that are not already bytes and parallelize with a thread pool.
  """
  with concurrent.futures.ThreadPoolExecutor(max_workers=max_threads) as executor:
    futures = []
    for input in request.inputs:
      futures.append(executor.submit(download_input, input))
    for future in concurrent.futures.as_completed(futures):
      try:
        future.result()
      except Exception as e:
        logger.exception(f"Error downloading input: {e}")



import logging
import os
from collections import namedtuple

import pytest

from clarifai.client import User
from clarifai.rag import RAG
from clarifai.urls.helper import ClarifaiUrlHelper

CREATE_APP_USER_ID = os.environ["CLARIFAI_USER_ID"]

TEXT_FILE_PATH = os.path.dirname(__file__) + "/assets/sample.txt"
PDF_URL = "https://samples.clarifai.com/test_doc.pdf"

auth_obj = namedtuple("auth", "ui")


@pytest.mark.requires_secrets
class TestRAG:

  @classmethod
  def setup_class(self):
    self.rag = RAG.setup(user_id=CREATE_APP_USER_ID)
    wf = self.rag._prompt_workflow
    auth = auth_obj(ui="https://clarifai.com")
    self.workflow_url = ClarifaiUrlHelper(auth).clarifai_url(wf.user_id, wf.app_id, "workflows",
                                                             wf.id)

  def test_setup_correct(self):
    assert len(self.rag._prompt_workflow.workflow_info.nodes) == 2

  def test_from_existing_workflow(self):
    agent = RAG(workflow_url=self.workflow_url)
    assert agent._app.id == self.rag._app.id

  def test_predict_client_manage_state(self):
    messages = [{"role": "human", "content": "What is 1 + 1?"}]
    new_messages = self.rag.chat(messages, client_manage_state=True)
    assert len(new_messages) == 2

  @pytest.mark.skip(reason="Not yet supported. Work in progress.")
  def test_predict_server_manage_state(self):
    messages = [{"role": "human", "content": "What is 1 + 1?"}]
    new_messages = self.rag.chat(messages)
    assert len(new_messages) == 1

  def test_upload_docs_filepath(self, caplog):
    with caplog.at_level(logging.INFO):
      self.rag.upload(file_path=TEXT_FILE_PATH)
      assert "SUCCESS" in caplog.text

  def test_upload_docs_from_url(self, caplog):
    with caplog.at_level(logging.INFO):
      self.rag.upload(url=PDF_URL)
      assert "SUCCESS" in caplog.text

  @classmethod
  def teardown_class(self):
    User(user_id=CREATE_APP_USER_ID).delete_app(self.rag._app.id)



from clarifai_grpc.grpc.api import resources_pb2

from clarifai.client.base import BaseClient
from clarifai.client.lister import Lister
from clarifai.utils.logging import logger


class Deployment(Lister, BaseClient):
  """Deployment is a class that provides access to Clarifai API endpoints related to Deployment information."""

  def __init__(self,
               deployment_id: str = None,
               user_id: str = None,
               base_url: str = "https://api.clarifai.com",
               pat: str = None,
               token: str = None,
               root_certificates_path: str = None,
               **kwargs):
    """Initializes a Deployment object.

    Args:
        deployment_id (str): The Deployment ID for the Deployment to interact with.
        user_id (str): The user ID of the user.
        base_url (str): Base API url. Default "https://api.clarifai.com"
        pat (str): A personal access token for authentication. Can be set as env var CLARIFAI_PAT
        token (str): A session token for authentication. Accepts either a session token or a pat. Can be set as env var CLARIFAI_SESSION_TOKEN
        root_certificates_path (str): Path to the SSL root certificates file, used to establish secure gRPC connections.
        **kwargs: Additional keyword arguments to be passed to the deployment.
    """
    self.kwargs = {**kwargs, 'id': deployment_id, 'user_id': user_id}
    self.deployment_info = resources_pb2.Deployment(**self.kwargs)
    self.logger = logger
    BaseClient.__init__(
        self,
        user_id=user_id,
        base=base_url,
        pat=pat,
        token=token,
        root_certificates_path=root_certificates_path)
    Lister.__init__(self)

  @staticmethod
  def get_runner_selector(user_id: str, deployment_id: str) -> resources_pb2.RunnerSelector:
    """Returns a RunnerSelector object for the given deployment_id.

    Args:
        deployment_id (str): The deployment ID for the deployment.

    Returns:
        resources_pb2.RunnerSelector: A RunnerSelector object for the given deployment_id.
    """
    return resources_pb2.RunnerSelector(deployment_id=deployment_id, user_id=user_id)

  def __getattr__(self, name):
    return getattr(self.deployment_info, name)

  def __str__(self):
    init_params = [param for param in self.kwargs.keys()]
    attribute_strings = [
        f"{param}={getattr(self.deployment_info, param)}" for param in init_params
        if hasattr(self.deployment_info, param)
    ]
    return f"Deployment Details: \n{', '.join(attribute_strings)}\n"



import click
from clarifai.cli.base import cli


@cli.group()
def model():
  """Manage models: upload, test locally"""
  pass


@model.command()
@click.option(
    '--model_path',
    type=click.Path(exists=True),
    required=True,
    help='Path to the model directory.')
@click.option(
    '--download_checkpoints',
    is_flag=True,
    help=
    'Flag to download checkpoints before uploading and including them in the tar file that is uploaded. Defaults to False, which will attempt to download them at docker build time.',
)
@click.option(
    '--skip_dockerfile',
    is_flag=True,
    help=
    'Flag to skip generating a dockerfile so that you can manually edit an already created dockerfile.',
)
def upload(model_path, download_checkpoints, skip_dockerfile):
  """Upload a model to Clarifai."""
  from clarifai.runners.models import model_upload

  model_upload.main(model_path, download_checkpoints, skip_dockerfile)


@model.command()
@click.option(
    '--model_path',
    type=click.Path(exists=True),
    required=True,
    help='Path to the model directory.')
def test_locally(model_path):
  """Test model locally."""
  try:
    from clarifai.runners.models import model_run_locally
    model_run_locally.main(model_path)
    click.echo(f"Model tested locally from {model_path}.")
  except Exception as e:
    click.echo(f"Failed to test model locally: {e}", err=True)


@model.command()
@click.option(
    '--model_path',
    type=click.Path(exists=True),
    required=True,
    help='Path to the model directory.')
def run_locally(model_path):
  """Run model locally and starts a GRPC server to serve the model."""
  try:
    from clarifai.runners.models import model_run_locally
    model_run_locally.main(model_path, run_model_server=True)
    click.echo(f"Model server started locally from {model_path}.")
  except Exception as e:
    click.echo(f"Failed to starts model server locally: {e}", err=True)



import datetime
import json
import logging

from rich.logging import RichHandler

from clarifai.utils.logging import JsonFormatter, _get_library_name, get_logger, set_logger_context


def test_get_logger():
  logger = get_logger("DEBUG", "test_logger")
  assert logger.level == logging.DEBUG
  assert logger.name == "test_logger"
  assert isinstance(logger.handlers[0], RichHandler)


def test_get_logger_defaults():
  logger = get_logger()
  assert logger.level == logging.NOTSET
  assert logger.name == _get_library_name()
  assert isinstance(logger.handlers[0], RichHandler)


def test_json_logger():
  filename = "testy.py"
  msg = "testy"
  lineno = 1
  sinfo = "testf2"
  r = logging.LogRecord(
      name="testy",
      level=logging.INFO,
      pathname=filename,
      lineno=lineno,
      msg=msg,
      args=(),
      exc_info=None,
      func="testf",
      sinfo=sinfo)
  jf = JsonFormatter()
  # format the record as a json line.
  json_line = jf.format(r)
  result = json.loads(json_line)
  # parse timestamp of format "2024-09-24T22:06:49.573038Z" in @timestamp field.
  assert result["@timestamp"] is not None
  ts = result["@timestamp"]
  # assert the ts was within 10 seconds of now (in UTC time)
  assert abs(datetime.datetime.utcnow() - datetime.datetime.strptime(
      ts, "%Y-%m-%dT%H:%M:%S.%fZ")) < datetime.timedelta(seconds=10)
  assert result['filename'] == filename
  assert result['msg'] == msg
  assert result['lineno'] == lineno
  assert result['stack_info'] == sinfo
  assert result['level'] == "info"
  assert 'req_id' not in result

  req_id = "test_req_id"
  set_logger_context(req_id=req_id)
  json_line = jf.format(r)
  result = json.loads(json_line)
  assert abs(datetime.datetime.utcnow() - datetime.datetime.strptime(
      ts, "%Y-%m-%dT%H:%M:%S.%fZ")) < datetime.timedelta(seconds=10)
  assert result['filename'] == filename
  assert result['msg'] == msg
  assert result['lineno'] == lineno
  assert result['stack_info'] == sinfo
  assert result['level'] == "info"
  assert result['req_id'] == req_id



from abc import ABC, abstractmethod
from typing import Iterator

from clarifai_grpc.grpc.api import service_pb2


class ModelClass(ABC):

  def predict_wrapper(
      self, request: service_pb2.PostModelOutputsRequest) -> service_pb2.MultiOutputResponse:
    """This method is used for input/output proto data conversion"""
    return self.predict(request)

  def generate_wrapper(self, request: service_pb2.PostModelOutputsRequest
                      ) -> Iterator[service_pb2.MultiOutputResponse]:
    """This method is used for input/output proto data conversion and yield outcome"""
    return self.generate(request)

  def stream_wrapper(self, request: service_pb2.PostModelOutputsRequest
                    ) -> Iterator[service_pb2.MultiOutputResponse]:
    """This method is used for input/output proto data conversion and yield outcome"""
    return self.stream(request)

  @abstractmethod
  def load_model(self):
    raise NotImplementedError("load_model() not implemented")

  @abstractmethod
  def predict(self,
              request: service_pb2.PostModelOutputsRequest) -> service_pb2.MultiOutputResponse:
    raise NotImplementedError("run_input() not implemented")

  @abstractmethod
  def generate(self, request: service_pb2.PostModelOutputsRequest
              ) -> Iterator[service_pb2.MultiOutputResponse]:
    raise NotImplementedError("generate() not implemented")

  @abstractmethod
  def stream(self, request_iterator: Iterator[service_pb2.PostModelOutputsRequest]
            ) -> Iterator[service_pb2.MultiOutputResponse]:
    raise NotImplementedError("stream() not implemented")



import importlib.util
import inspect
import os
import shutil
import subprocess
import sys
import tempfile
import traceback
import venv

from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2, status_pb2
from clarifai_protocol import BaseRunner

from clarifai.runners.models.model_upload import ModelUploader
from clarifai.utils.logging import logger


class ModelRunLocally:

  def __init__(self, model_path):
    self.model_path = model_path
    self.requirements_file = os.path.join(self.model_path, "requirements.txt")

  def create_temp_venv(self):
    """Create a temporary virtual environment."""
    logger.info("Creating temporary virtual environment...")
    temp_dir = tempfile.mkdtemp()
    venv_dir = os.path.join(temp_dir, "venv")
    venv.create(venv_dir, with_pip=True)

    self.venv_dir = venv_dir
    self.temp_dir = temp_dir
    self.python_executable = os.path.join(venv_dir, "bin", "python")

    logger.info(f"Created temporary virtual environment at {venv_dir}")
    return venv_dir, temp_dir

  def install_requirements(self):
    """Install the dependencies from requirements.txt and Clarifai."""
    pip_executable = os.path.join(self.venv_dir, "bin", "pip")
    try:
      logger.info(
          f"Installing requirements from {self.requirements_file}... in the virtual environment {self.venv_dir}"
      )
      subprocess.check_call([pip_executable, "install", "-r", self.requirements_file])
      logger.info("Installing Clarifai package...")
      subprocess.check_call([pip_executable, "install", "clarifai"])
      logger.info("Requirements installed successfully!")
    except subprocess.CalledProcessError as e:
      logger.error(f"Error installing requirements: {e}")
      self.clean_up()
      sys.exit(1)

  def _get_model_runner(self):
    """Dynamically import the runner class from the model file."""

    # import the runner class that to be implement by the user
    runner_path = os.path.join(self.model_path, "1", "model.py")

    # arbitrary name given to the module to be imported
    module = "runner_module"

    spec = importlib.util.spec_from_file_location(module, runner_path)
    runner_module = importlib.util.module_from_spec(spec)
    sys.modules[module] = runner_module
    spec.loader.exec_module(runner_module)

    # Find all classes in the model.py file that are subclasses of BaseRunner
    classes = [
        cls for _, cls in inspect.getmembers(runner_module, inspect.isclass)
        if issubclass(cls, BaseRunner) and cls.__module__ == runner_module.__name__
    ]

    #  Ensure there is exactly one subclass of BaseRunner in the model.py file
    if len(classes) != 1:
      raise Exception("Expected exactly one subclass of BaseRunner, found: {}".format(
          len(classes)))

    MyRunner = classes[0]
    return MyRunner

  def _build_request(self):
    """Create a mock inference request for testing the model."""

    uploader = ModelUploader(self.model_path)
    model_version_proto = uploader.get_model_version_proto()
    model_version_proto.id = "model_version"

    return service_pb2.PostModelOutputsRequest(
        model=resources_pb2.Model(model_version=model_version_proto),
        inputs=[
            resources_pb2.Input(data=resources_pb2.Data(
                text=resources_pb2.Text(raw="How many people live in new york?"),
                image=resources_pb2.Image(url="https://samples.clarifai.com/metro-north.jpg"),
                audio=resources_pb2.Audio(url="https://samples.clarifai.com/GoodMorning.wav"),
            ))
        ],
    )

  def _run_model_inference(self, runner):
    """Perform inference using the runner."""
    request = self._build_request()

    try:
      return runner.predict(request)
    except Exception as e:
      logger.error(f"Model Prediction failed: {e}")
      traceback.print_exc()
      return service_pb2.MultiOutputResponse(status=status_pb2.Status(
          code=status_code_pb2.MODEL_PREDICTION_FAILED,
          description="Prediction failed",
          details="",
          internal_details=str(e),
      ))

  def _run_test(self):
    """Test the model locally by making a prediction."""
    # validate that we have checkpoints downloaded before constructing MyRunner
    uploader = ModelUploader(self.model_path)
    uploader.download_checkpoints()
    # construct MyRunner which will call load_model()
    MyRunner = self._get_model_runner()
    runner = MyRunner(
        runner_id="n/a",
        nodepool_id="n/a",
        compute_cluster_id="n/a",
        user_id="n/a",
    )

    # send an inference.
    response = self._run_model_inference(runner)
    if response.outputs[0].status.code != status_code_pb2.SUCCESS:
      logger.error(f"Moddel Prediction failed: {response}")
    else:
      logger.info(f"Model Prediction succeeded: {response}")

  def test_model(self):
    """Test the model by running it locally in the virtual environment."""
    command = [
        self.python_executable,
        "-c",
        f"import sys; sys.path.append('{os.path.dirname(os.path.abspath(__file__))}'); "
        f"from model_run_locally import ModelRunLocally; ModelRunLocally('{self.model_path}')._run_test()",
    ]
    try:
      logger.info("Testing the model locally...")
      subprocess.check_call(command)
      logger.info("Model tested successfully!")
    except subprocess.CalledProcessError as e:
      logger.error(f"Error testing the model: {e}")
      sys.exit(1)

  # run the model server
  def run_model_server(self):
    """Run the Clarifai Runners's model server."""

    command = [
        self.python_executable, "-m", "clarifai.runners.server", "--model_path", self.model_path,
        "--start_dev_server"
    ]
    try:
      logger.info(f"Starting model server with model at {self.model_path}...")
      subprocess.check_call(command)
      logger.info("Model server started successfully!")
    except subprocess.CalledProcessError as e:
      logger.error(f"Error running model server: {e}")
      self.clean_up()
      sys.exit(1)

  def clean_up(self):
    """Clean up the temporary virtual environment."""
    if os.path.exists(self.temp_dir):
      logger.info("Cleaning up temporary virtual environment...")
      shutil.rmtree(self.temp_dir)


def main(model_path, run_model_server=False):

  manager = ModelRunLocally(model_path)
  manager.create_temp_venv()

  try:
    manager.install_requirements()
    if run_model_server:
      manager.run_model_server()
    else:
      manager.test_model()
  finally:
    manager.clean_up()



from math import ceil
from typing import Any, Callable, Dict, Generator

from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2
from google.protobuf.json_format import MessageToDict
from google.protobuf.struct_pb2 import Struct
from schema import SchemaError

from clarifai.client.base import BaseClient
from clarifai.client.input import Inputs
from clarifai.client.lister import Lister
from clarifai.constants.search import (DEFAULT_SEARCH_ALGORITHM, DEFAULT_SEARCH_METRIC,
                                       DEFAULT_TOP_K)
from clarifai.errors import UserError
from clarifai.schema.search import get_schema


class Search(Lister, BaseClient):

  def __init__(self,
               user_id: str,
               app_id: str,
               top_k: int = None,
               metric: str = DEFAULT_SEARCH_METRIC,
               algorithm: str = DEFAULT_SEARCH_ALGORITHM,
               pagination: bool = False,
               base_url: str = "https://api.clarifai.com",
               pat: str = None,
               token: str = None,
               root_certificates_path: str = None):
    """Initialize the Search object.

    Args:
        user_id (str): User ID.
        app_id (str): App ID.
        top_k (int, optional): Top K results to retrieve. Defaults to 10.
        metric (str, optional): Similarity metric (either 'cosine' or 'euclidean'). Defaults to 'cosine'.
        alogrithm (str, optional): Search algorithm (either 'nearest_neighbor' or 'brute_force'). Defaults to 'nearest_neighbor'.
        pagination (bool, optional): Enable pagination. Defaults to False.
        base_url (str, optional): Base API url. Defaults to "https://api.clarifai.com".
        pat (str, optional): A personal access token for authentication. Can be set as env var CLARIFAI_PAT
        token (str): A session token for authentication. Accepts either a session token or a pat. Can be set as env var CLARIFAI_SESSION_TOKEN
        root_certificates_path (str): Path to the SSL root certificates file, used to establish secure gRPC connections.

    Raises:
        UserError: If the metric is not 'cosine' or 'euclidean'.
        UserError: If the algorithm is not 'nearest_neighbor' or 'brute_force'.
    """
    if metric not in ["cosine", "euclidean"]:
      raise UserError("Metric should be either cosine or euclidean")
    if algorithm not in ["nearest_neighbor", "brute_force"]:
      raise UserError("Algorithm should be either nearest_neighbor or brute_force")
    if metric == "cosine" and algorithm == "nearest_neighbor":
      raise UserError("Cosine distance metric is not supported with nearest neighbor algorithm")
    if top_k and pagination:
      raise UserError(
          "top_k and pagination cannot be used together. Please set pagination to False.")
    if not top_k and not pagination:
      top_k = DEFAULT_TOP_K

    self.user_id = user_id
    self.app_id = app_id
    self.metric_distance = dict(cosine="COSINE_DISTANCE", euclidean="EUCLIDEAN_DISTANCE")[metric]
    self.algorithm = algorithm
    self.data_proto = resources_pb2.Data()
    self.top_k = top_k
    self.pagination = pagination
    self.inputs = Inputs(
        user_id=self.user_id, app_id=self.app_id, pat=pat, token=token, base_url=base_url)
    self.rank_filter_schema = get_schema()
    BaseClient.__init__(
        self,
        user_id=self.user_id,
        app_id=self.app_id,
        base=base_url,
        pat=pat,
        token=token,
        root_certificates_path=root_certificates_path)
    Lister.__init__(self, page_size=1000)

  def _get_annot_proto(self, **kwargs):
    """Get an Annotation proto message based on keyword arguments.

    Args:
        **kwargs: Keyword arguments specifying the resource.

    Returns:
        resources_pb2.Annotation: An Annotation proto message.
    """
    if not kwargs:
      return resources_pb2.Annotation()

    self.data_proto = resources_pb2.Data()
    for key, value in kwargs.items():
      if key == "image_bytes":
        image_proto = self.inputs.get_input_from_bytes("", image_bytes=value).data.image
        self.data_proto.image.CopyFrom(image_proto)

      elif key == "image_url":
        image_proto = self.inputs.get_input_from_url("", image_url=value).data.image
        self.data_proto.image.CopyFrom(image_proto)

      elif key == "concepts":
        for concept in value:
          concept_proto = resources_pb2.Concept(**concept)
          self.data_proto.concepts.add().CopyFrom(concept_proto)

      elif key == "text_raw":
        text_proto = self.inputs.get_input_from_bytes(
            "", text_bytes=bytes(value, 'utf-8')).data.text
        self.data_proto.text.CopyFrom(text_proto)

      elif key == "metadata":
        metadata_struct = Struct()
        metadata_struct.update(value)
        self.data_proto.metadata.CopyFrom(metadata_struct)

      elif key == "geo_point":
        geo_point_proto = self._get_geo_point_proto(value["longitude"], value["latitude"],
                                                    value["geo_limit"])
        self.data_proto.geo.CopyFrom(geo_point_proto)

      else:
        raise UserError(f"kwargs contain key that is not supported: {key}")
    return resources_pb2.Annotation(data=self.data_proto)

  def _get_input_proto(self, **kwargs):
    """Get an Input proto message based on keyword arguments.

    Args:
        **kwargs: Keyword arguments specifying the resource.

    Returns:
        resources_pb2.Input: An Input proto message.
    """
    if not kwargs:
      return resources_pb2.Input()

    self.input_proto = resources_pb2.Input()
    self.data_proto = resources_pb2.Data()
    for key, value in kwargs.items():
      if key == "input_types":
        for input_type in value:
          if input_type == "image":
            self.data_proto.image.CopyFrom(resources_pb2.Image())
          elif input_type == "text":
            self.data_proto.text.CopyFrom(resources_pb2.Text())
          elif input_type == "audio":
            self.data_proto.audio.CopyFrom(resources_pb2.Audio())
          elif input_type == "video":
            self.data_proto.video.CopyFrom(resources_pb2.Video())
        self.input_proto.data.CopyFrom(self.data_proto)
      elif key == "input_dataset_ids":
        self.input_proto.dataset_ids.extend(value)
      elif key == "input_status_code":
        self.input_proto.status.code = value
      else:
        raise UserError(f"kwargs contain key that is not supported: {key}")
    return self.input_proto

  def _get_geo_point_proto(self, longitude: float, latitude: float,
                           geo_limit: float) -> resources_pb2.Geo:
    """Get a GeoPoint proto message based on geographical data.

    Args:
        longitude (float): Longitude coordinate.
        latitude (float): Latitude coordinate.
        geo_limit (float): Geographical limit.

    Returns:
        resources_pb2.Geo: A Geo proto message.
    """
    return resources_pb2.Geo(
        geo_point=resources_pb2.GeoPoint(longitude=longitude, latitude=latitude),
        geo_limit=resources_pb2.GeoLimit(type="withinKilometers", value=geo_limit))

  def _list_topk_generator(self, endpoint: Callable[..., Any], proto_message: Any,
                           request_data: Dict[str, Any]) -> Generator[Dict[str, Any], None, None]:
    """Lists all pages of a resource.

        Args:
            endpoint (Callable): The endpoint to call.
            proto_message (Any): The proto message to use.
            request_data (dict): The request data to use.

        Yields:
            response_dict: The next item in the listing.
        """
    max_pages = ceil(self.top_k / self.default_page_size)
    total_hits = 0
    page = 1
    while (page <= max_pages):
      if (page == max_pages):
        per_page = self.top_k - total_hits
      else:
        per_page = self.default_page_size
      request_data['pagination'] = service_pb2.Pagination(page=page, per_page=per_page)
      response = self._grpc_request(endpoint, proto_message(**request_data))
      dict_response = MessageToDict(response, preserving_proto_field_name=True)
      if response.status.code != status_code_pb2.SUCCESS:
        if "page * perPage cannot exceed" in str(response.status.details):
          msg = (f"Your top_k is set to {self.top_k}. "
                 f"The current pagination settings exceed the limit. Please reach out to "
                 f"support@clarifai.com to request an increase for your use case.\n"
                 f"req_id: {response.status.req_id}")
          raise UserError(msg)
        else:
          raise Exception(f"Listing failed with response {response!r}")

      if 'hits' not in list(dict_response.keys()):
        break
      page += 1
      total_hits += per_page
      yield response

  def _list_all_pages_generator(self,
                                endpoint: Callable,
                                proto_message: Any,
                                request_data: Dict[str, Any],
                                page_no: int = None,
                                per_page: int = None) -> Generator[Dict[str, Any], None, None]:
    """Lists pages of a resource.

    Args:
        endpoint (Callable): The endpoint to call.
        proto_message (Any): The proto message to use.
        request_data (dict): The request data to use.
        page_no (int): The page number to list.
        per_page (int): The number of items per page.

    Yields:
        response_dict: The next item in the listing.
    """
    page = 1 if not page_no else page_no
    if page_no and not per_page:
      per_page = self.default_page_size
    while True:
      request_data['pagination'] = service_pb2.Pagination(page=page, per_page=per_page)
      response = self._grpc_request(endpoint, proto_message(**request_data))
      dict_response = MessageToDict(response, preserving_proto_field_name=True)
      if response.status.code != status_code_pb2.SUCCESS:
        if "page * perPage cannot exceed" in str(response.status.details):
          msg = (f"Your pagination is set to {page_no*per_page}. "
                 f"The current pagination settings exceed the limit. Please reach out to "
                 f"support@clarifai.com to request an increase for your use case.\n"
                 f"req_id: {response.status.req_id}")
          raise UserError(msg)
        else:
          raise Exception(f"Listing failed with response {response!r}")
      if 'hits' not in list(dict_response.keys()):
        break
      yield response
      if page_no is not None or per_page is not None:
        break
      page += 1

  def query(
      self,
      ranks=[{}],
      filters=[{}],
      page_no: int = None,
      per_page: int = None,
  ):
    """Perform a query with rank and filters.

    Args:
        ranks (List[Dict], optional): List of rank parameters. Defaults to [{}].
        filters (List[Dict], optional): List of filter parameters. Defaults to [{}].
        page_no (int): The page number to list.
        per_page (int): The number of items per page.

    Returns:
        Generator[Dict[str, Any], None, None]: A generator of query results.

    Examples:
        Get successful inputs of type image or text
        >>> from clarifai.client.search import Search
        >>> search = Search(user_id='user_id', app_id='app_id', top_k=10, metric='cosine')
        # This performs OR operation on input_types and input_status_code
        >>> res = search.query(filters=[{'input_types': ['image', 'text'], 'input_status_code': 30000}])

        Vector search over inputs
        >>> from clarifai.client.search import Search
        >>> search = Search(user_id='user_id', app_id='app_id' , metric='cosine', pagination=True)
        >>> res = search.query(ranks=[{'image_url': 'https://samples.clarifai.com/dog.tiff'}],page_no=2, per_page=5)

    Note:
        For schema of rank and filter, please refer to [schema](https://github.com/Clarifai/clarifai-python/tree/master/clarifai/schema/search.py).
        For more detailed search examples, please refer to [examples](https://github.com/Clarifai/examples/tree/main/search).
    """
    if not self.pagination and (per_page or page_no):
      raise UserError("Pagination settings are only available when pagination is enabled."
                      "Please set Search(pagination=True) while initializing Search().")
    try:
      self.rank_filter_schema.validate(ranks)
      self.rank_filter_schema.validate(filters)
    except SchemaError as err:
      raise UserError(f"Invalid rank or filter input: {err}")

    # For each rank, create a Rank proto message
    # For ranks it only allows resources_pb2.Annotation proto, so no need of splitting protos into annotation and input.
    all_ranks = []
    for rank_dict in ranks:
      rank_annot_proto = self._get_annot_proto(**rank_dict)
      all_ranks.append(resources_pb2.Rank(annotation=rank_annot_proto))

    all_filters = []
    # check for filters which is compatible with input proto
    for each_filter in filters:
      input_dict = {
          key: each_filter.pop(key)
          for key in ['input_types', 'input_dataset_ids', 'input_status_code']
          if key in each_filter
      }

      all_filters.append(
          resources_pb2.Filter(
              annotation=self._get_annot_proto(**each_filter),
              input=self._get_input_proto(**input_dict)))

    # Create a PostInputsSearchesRequest proto message
    request_data = dict(
        user_app_id=self.user_app_id,
        searches=[
            resources_pb2.Search(
                query=resources_pb2.Query(ranks=all_ranks, filters=all_filters),
                algorithm=self.algorithm,
                metric=self.metric_distance)
        ])
    # Calls PostInputsSearches for annotation ranks, input filters
    if self.pagination:
      return self._list_all_pages_generator(self.STUB.PostInputsSearches,
                                            service_pb2.PostInputsSearchesRequest, request_data,
                                            page_no, per_page)
    return self._list_topk_generator(self.STUB.PostInputsSearches,
                                     service_pb2.PostInputsSearchesRequest, request_data)



import os
from concurrent.futures import ThreadPoolExecutor
from typing import List, Tuple, Type

from clarifai_grpc.grpc.api import resources_pb2
from google.protobuf.struct_pb2 import Struct

from clarifai.client.input import Inputs
from clarifai.datasets.upload.base import ClarifaiDataLoader, ClarifaiDataset
from clarifai.utils.misc import get_uuid


class VisualClassificationDataset(ClarifaiDataset):

  def __init__(self,
               data_generator: Type[ClarifaiDataLoader],
               dataset_id: str,
               max_workers: int = 4) -> None:
    super().__init__(data_generator, dataset_id, max_workers)

  def _extract_protos(self, batch_input_ids: List[str]
                     ) -> Tuple[List[resources_pb2.Input], List[resources_pb2.Annotation]]:
    """Create input image and annotation protos for batch of input ids.
    Args:
      batch_input_ids: List of input IDs to retrieve the protos for.
    Returns:
      input_protos: List of input protos.
      annotation_protos: List of annotation protos.
    """
    input_protos, annotation_protos = [], []

    def process_data_item(id):
      data_item = self.data_generator[id]
      metadata = Struct()
      image_path = data_item.image_path
      labels = data_item.labels if isinstance(data_item.labels,
                                              list) else [data_item.labels]  # clarifai concept
      label_ids = data_item.label_ids
      input_id = f"{self.dataset_id}-{get_uuid(8)}" if data_item.id is None else f"{self.dataset_id}-{str(data_item.id)}"
      geo_info = data_item.geo_info
      if data_item.metadata is not None:
        metadata.update(data_item.metadata)
      elif image_path is not None:
        metadata.update({"filename": os.path.basename(image_path)})
      else:
        metadata = None

      self.all_input_ids[id] = input_id
      if data_item.image_bytes is not None:
        input_protos.append(
            Inputs.get_input_from_bytes(
                input_id=input_id,
                image_bytes=data_item.image_bytes,
                dataset_id=self.dataset_id,
                labels=labels,
                label_ids=label_ids,
                geo_info=geo_info,
                metadata=metadata))
      else:
        input_protos.append(
            Inputs.get_input_from_file(
                input_id=input_id,
                image_file=image_path,
                dataset_id=self.dataset_id,
                labels=labels,
                label_ids=label_ids,
                geo_info=geo_info,
                metadata=metadata))

    with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
      futures = [executor.submit(process_data_item, id) for id in batch_input_ids]
      for job in futures:
        job.result()

    return input_protos, annotation_protos


class VisualDetectionDataset(ClarifaiDataset):
  """Visual detection dataset proto class."""

  def __init__(self,
               data_generator: Type[ClarifaiDataLoader],
               dataset_id: str,
               max_workers: int = 4) -> None:
    super().__init__(data_generator, dataset_id, max_workers)

  def _extract_protos(self, batch_input_ids: List[int]
                     ) -> Tuple[List[resources_pb2.Input], List[resources_pb2.Annotation]]:
    """Create input image protos for each data generator item.
    Args:
      batch_input_ids: List of input IDs to retrieve the protos for.
    Returns:
      input_protos: List of input protos.
      annotation_protos: List of annotation protos.
    """
    input_protos, annotation_protos = [], []

    def process_data_item(id):
      data_item = self.data_generator[id]
      metadata = Struct()
      image = data_item.image_path
      labels = data_item.labels  # list:[l1,...,ln]
      if data_item.label_ids is not None:
        assert len(labels) == len(
            data_item.label_ids), "Length of labels and label_ids must be equal"
        label_ids = data_item.label_ids
      else:
        label_ids = None
      bboxes = data_item.bboxes  # [[xmin,ymin,xmax,ymax],...,[xmin,ymin,xmax,ymax]]
      input_id = f"{self.dataset_id}-{get_uuid(8)}" if data_item.id is None else f"{self.dataset_id}-{str(data_item.id)}"
      if data_item.metadata is not None:
        metadata.update(data_item.metadata)
      else:
        metadata.update({"filename": os.path.basename(image)})
      geo_info = data_item.geo_info

      self.all_input_ids[id] = input_id
      if data_item.image_bytes is not None:
        input_protos.append(
            Inputs.get_input_from_bytes(
                input_id=input_id,
                image_bytes=data_item.image_bytes,
                dataset_id=self.dataset_id,
                geo_info=geo_info,
                metadata=metadata))
      else:
        input_protos.append(
            Inputs.get_input_from_file(
                input_id=input_id,
                image_file=image,
                dataset_id=self.dataset_id,
                geo_info=geo_info,
                metadata=metadata))
      # iter over bboxes and labels
      # one id could have more than one bbox and label
      for i in range(len(bboxes)):
        annotation_protos.append(
            Inputs.get_bbox_proto(
                input_id=input_id,
                label=labels[i],
                bbox=bboxes[i],
                label_id=label_ids[i] if label_ids else None))

    with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
      futures = [executor.submit(process_data_item, id) for id in batch_input_ids]
      for job in futures:
        job.result()

    return input_protos, annotation_protos


class VisualSegmentationDataset(ClarifaiDataset):
  """Visual segmentation dataset proto class."""

  def __init__(self,
               data_generator: Type[ClarifaiDataLoader],
               dataset_id: str,
               max_workers: int = 4) -> None:
    super().__init__(data_generator, dataset_id, max_workers)

  def _extract_protos(self, batch_input_ids: List[str]
                     ) -> Tuple[List[resources_pb2.Input], List[resources_pb2.Annotation]]:
    """Create input image and annotation protos for batch of input ids.
    Args:
      batch_input_ids: List of input IDs to retrieve the protos for.
    Returns:
      input_protos: List of input protos.
      annotation_protos: List of annotation protos.
    """
    input_protos, annotation_protos = [], []

    def process_data_item(id):
      data_item = self.data_generator[id]
      metadata = Struct()
      image = data_item.image_path
      labels = data_item.labels
      if data_item.label_ids is not None:
        assert len(labels) == len(
            data_item.label_ids), "Length of labels and label_ids must be equal"
        label_ids = data_item.label_ids
      else:
        label_ids = None
      _polygons = data_item.polygons  # list of polygons: [[[x,y],...,[x,y]],...]
      input_id = f"{self.dataset_id}-{get_uuid(8)}" if data_item.id is None else f"{self.dataset_id}-{str(data_item.id)}"
      if data_item.metadata is not None:
        metadata.update(data_item.metadata)
      else:
        metadata.update({"filename": os.path.basename(image)})
      geo_info = data_item.geo_info

      self.all_input_ids[id] = input_id
      if data_item.image_bytes is not None:
        input_protos.append(
            Inputs.get_input_from_bytes(
                input_id=input_id,
                image_bytes=data_item.image_bytes,
                dataset_id=self.dataset_id,
                geo_info=geo_info,
                metadata=metadata))
      else:
        input_protos.append(
            Inputs.get_input_from_file(
                input_id=input_id,
                image_file=image,
                dataset_id=self.dataset_id,
                geo_info=geo_info,
                metadata=metadata))

      ## Iterate over each masked image and create a proto for upload to clarifai
      ## The length of masks/polygons-list and labels must be equal
      for i, _polygon in enumerate(_polygons):
        try:
          annotation_protos.append(
              Inputs.get_mask_proto(
                  input_id=input_id,
                  label=labels[i],
                  polygons=_polygon,
                  label_id=label_ids[i] if label_ids else None))
        except IndexError:
          continue

    with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
      futures = [executor.submit(process_data_item, id) for id in batch_input_ids]
      for job in futures:
        job.result()

    return input_protos, annotation_protos



import json
import os
import tempfile
import zipfile
from concurrent.futures import ThreadPoolExecutor, as_completed
from io import BytesIO
from typing import Any, Dict, Iterator, List, Optional

import requests
from clarifai_grpc.grpc.api import resources_pb2
from google.protobuf.json_format import MessageToDict
from PIL import ImageFile
from tqdm import tqdm

from clarifai.constants.dataset import CONTENT_TYPE
from clarifai.errors import UserError
from clarifai.utils.logging import logger


class DatasetExportReader:

  def __init__(self,
               session: requests.Session = None,
               archive_url: Optional[str] = None,
               local_archive_path: Optional[str] = None):
    """Download/Reads the zipfile archive and yields every api.Input object.

    Args:
        session: requests.Session object
        archive_url: URL of the DatasetVersionExport archive
        local_archive_path: Path to the DatasetVersionExport archive
    """
    self.input_count = None
    self.temp_file = None
    self.session = session
    if not self.session:
      self.session = requests.Session()

    assert archive_url or local_archive_path, UserError(
        "Either archive_url or local_archive_path must be provided.")
    assert not (archive_url and local_archive_path), UserError(
        "Only one of archive_url or local_archive_path must be provided.")

    if archive_url:
      logger.info('url: %s' % archive_url)
      self.temp_file = self._download_temp_archive(archive_url)
      self.archive = zipfile.ZipFile(self.temp_file)
    else:
      logger.info("path: %s" % local_archive_path)
      self.archive = zipfile.ZipFile(local_archive_path)

    self.file_name_list = self.archive.namelist()
    assert "mimetype" in self.file_name_list, "Missing mimetype file in the dataset export archive."
    assert self.archive.read("mimetype") == b"application/x.clarifai-data+protobuf"
    self.file_name_list.remove("mimetype")

    logger.info("Obtained file name list. %d entries." % len(self.file_name_list))
    self.split_dir = os.path.dirname(self.file_name_list[0]) if len(self.file_name_list) else ""

  def _download_temp_archive(self, archive_url: str,
                             chunk_size: int = 128) -> tempfile.TemporaryFile:
    """Downloads the temp archive of InputBatches."""
    r = self.session.get(archive_url, stream=True)
    if r.headers['content-type'] == CONTENT_TYPE['json']:
      raise Exception("File is a json file :\n {}".format(r.json()))
    elif r.headers['content-type'] != CONTENT_TYPE['zip']:
      raise Exception('File is not a zip file')
    temp_file = tempfile.TemporaryFile()
    for chunk in r.iter_content(chunk_size=chunk_size):
      temp_file.write(chunk)

    return temp_file

  def __len__(self) -> int:
    if self.input_count is None:
      input_count = 0
      if self.file_name_list is not None:
        for filename in self.file_name_list:
          input_count += int(filename.split('_n')[-1])
      self.input_count = input_count

    return self.input_count

  def __iter__(self) -> Iterator[resources_pb2.Input]:
    """Loops through all InputBatches in the DatasetVersionExport and yields every api.Input object"""
    if self.file_name_list is not None:
      for filename in self.file_name_list:
        db = resources_pb2.InputBatch().FromString(self.archive.read(filename))
        for db_input in db.inputs:
          yield db_input

  def __enter__(self) -> 'DatasetExportReader':
    return self

  def __exit__(self, *args: Any) -> None:
    self.close()

  def close(self) -> None:
    logger.info("closing file objects.")
    self.archive.close()
    if self.temp_file:
      self.temp_file.close()


class InputAnnotationDownloader:

  def __init__(self,
               session: requests.Session,
               input_iterator: DatasetExportReader,
               num_workers: int = 4):
    """Downloads the archive from the URL into an archive of inputs, annotations in the directory format
    {split}/inputs and {split}/annotations.

    Args:
        session: requests.Session object
        input_iterator: Iterable of DatasetExportReader object
        num_workers: Number of threads to use for downloading
    """
    self.input_iterator = input_iterator
    self.num_workers = min(num_workers, 10)  # Max 10 threads
    self.num_inputs = 0
    self.num_annotations = 0
    self.split_prefix = None
    self.session = session
    self.input_ext = dict(image=".png", text=".txt", audio=".mp3", video=".mp4")
    if isinstance(self.input_iterator, DatasetExportReader):
      self.split_prefix = self.input_iterator.split_dir

  def _save_image_to_archive(self, new_archive: zipfile.ZipFile, hosted_url: str,
                             file_name: str) -> None:
    """Use PIL ImageFile to return image parsed from the response bytestring (from requests) and append to zip file."""
    p = ImageFile.Parser()
    p.feed(self.session.get(hosted_url).content)
    image = p.close()
    image_file = BytesIO()
    image.save(image_file, 'PNG')
    new_archive.writestr(file_name, image_file.getvalue())

  def _save_text_to_archive(self, new_archive: zipfile.ZipFile, hosted_url: str,
                            file_name: str) -> None:
    """Gets the text response bytestring (from requests) and append to zip file."""
    text_content = self.session.get(hosted_url).content
    new_archive.writestr(file_name, text_content)

  def _save_audio_to_archive(self, new_archive: zipfile.ZipFile, hosted_url: str,
                             file_name: str) -> None:
    """Gets the audio response bytestring (from requests) as chunks and append to zip file."""
    audio_response = self.session.get(hosted_url, stream=True)
    audio_stream = BytesIO()
    # Retrieve the audio content in chunks and write to the BytesIO object
    for chunk in audio_response.iter_content(chunk_size=128):
      audio_stream.write(chunk)
    new_archive.writestr(file_name, audio_stream.getvalue())

  def _save_video_to_archive(self, new_archive: zipfile.ZipFile, hosted_url: str,
                             file_name: str) -> None:
    """Gets the video response bytestring (from requests) as chunks and append to zip file."""
    video_response = self.session.get(hosted_url)
    video_stream = BytesIO()
    # Retrieve the video content in chunks and write to the BytesIO object
    for chunk in video_response.iter_content(chunk_size=128):
      video_stream.write(chunk)
    new_archive.writestr(file_name, video_stream.getvalue())

  def _save_annotation_to_archive(self, new_archive: zipfile.ZipFile, annot_data: List[Dict],
                                  file_name: str) -> None:
    """Gets the annotation response bytestring (from requests) and append to zip file."""
    # Serialize the dictionary to a JSON string
    json_str = json.dumps(annot_data)
    # Convert the JSON string to bytes
    bytes_object = json_str.encode()

    new_archive.writestr(file_name, bytes_object)

  def _write_archive(self, input_, new_archive, split: Optional[str] = None) -> None:
    """Writes the input, annotation archive into prefix dir."""
    data_dict = MessageToDict(input_.data)
    input_type = list(filter(lambda x: x in list(data_dict.keys()),
                             list(self.input_ext.keys())))[0]
    hosted = getattr(input_.data, input_type).hosted
    if hosted.prefix:
      assert 'orig' in hosted.sizes
      hosted_url = f"{hosted.prefix}/orig/{hosted.suffix}"
      file_name = os.path.join(split, "inputs", input_.id + self.input_ext[input_type])
      if input_type == "image":
        self._save_image_to_archive(new_archive, hosted_url, file_name)
      elif input_type == "text":
        self._save_text_to_archive(new_archive, hosted_url, file_name)
      elif input_type == "audio":
        self._save_audio_to_archive(new_archive, hosted_url, file_name)
      elif input_type == "video":
        self._save_video_to_archive(new_archive, hosted_url, file_name)
      self.num_inputs += 1

    if data_dict.get("metadata") or data_dict.get("concepts") or data_dict.get("regions"):
      file_name = os.path.join(split, "annotations", input_.id + ".json")
      annot_data = [{
          "metadata": data_dict.get("metadata", {})
      }] + data_dict.get("regions", []) + data_dict.get("concepts", [])

      self._save_annotation_to_archive(new_archive, annot_data, file_name)
      self.num_annotations += 1

  def _check_output_archive(self, save_path: str) -> None:
    try:
      archive = zipfile.ZipFile(save_path, 'r')
    except zipfile.BadZipFile as e:
      raise e
    assert len(
        archive.namelist()
    ) == self.num_inputs + self.num_annotations, "Archive has %d inputs+annotations | expecting %d inputs+annotations" % (
        len(archive.namelist()), self.num_inputs + self.num_annotations)

  def download_archive(self, save_path: str, split: Optional[str] = None) -> None:
    """Downloads the archive from the URL into an archive of inputs, annotations in the directory format
    {split}/inputs and {split}/annotations.
    """
    with zipfile.ZipFile(save_path, "a") as new_archive:
      with ThreadPoolExecutor(max_workers=self.num_workers) as executor:
        with tqdm(total=len(self.input_iterator), desc='Downloading Dataset') as progress:
          # Submit all jobs to the executor and store the returned futures
          futures = [
              executor.submit(self._write_archive, input_, new_archive, split)
              for input_ in self.input_iterator
          ]

          for _ in as_completed(futures):
            progress.update()

    self._check_output_archive(save_path)
    logger.info("Downloaded %d inputs and %d annotations to %s" %
                (self.num_inputs, self.num_annotations, save_path))



from typing import Iterator

from clarifai_grpc.grpc.api import service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2, status_pb2

from clarifai_protocol import BaseRunner
from clarifai_protocol.utils.health import HealthProbeRequestHandler
from ..utils.url_fetcher import ensure_urls_downloaded

from .model_class import ModelClass


class ModelRunner(BaseRunner, ModelClass, HealthProbeRequestHandler):
  """
  This is a subclass of the runner class which will handle only the work items relevant to models.

  It is also a subclass of ModelClass so that any subclass of ModelRunner will need to just
  implement predict(), generate() and stream() methods and load_model() if needed.
  """

  def __init__(
      self,
      runner_id: str,
      nodepool_id: str,
      compute_cluster_id: str,
      user_id: str = None,
      check_runner_exists: bool = True,
      base_url: str = "https://api.clarifai.com",
      pat: str = None,
      token: str = None,
      num_parallel_polls: int = 4,
      **kwargs,
  ) -> None:
    super().__init__(
        runner_id,
        nodepool_id,
        compute_cluster_id,
        user_id,
        check_runner_exists,
        base_url,
        pat,
        token,
        num_parallel_polls,
        **kwargs,
    )
    self.load_model()

    # After model load successfully set the health probe to ready and startup
    HealthProbeRequestHandler.is_ready = True
    HealthProbeRequestHandler.is_startup = True

  def get_runner_item_output_for_status(self,
                                        status: status_pb2.Status) -> service_pb2.RunnerItemOutput:
    """
    Set the error message in the RunnerItemOutput message subfield, used during exception handling
    where we may only have a status to return.

    Args:
      status: status_pb2.Status - the status to return

    Returns:
      service_pb2.RunnerItemOutput - the RunnerItemOutput message with the status set
    """
    rio = service_pb2.RunnerItemOutput(
        multi_output_response=service_pb2.MultiOutputResponse(status=status))
    return rio

  def runner_item_predict(self,
                          runner_item: service_pb2.RunnerItem) -> service_pb2.RunnerItemOutput:
    """
    Run the model on the given request. You shouldn't need to override this method, see run_input
    for the implementation to process each input in the request.

    Args:
      request: service_pb2.PostModelOutputsRequest - the request to run the model on

    Returns:
      service_pb2.MultiOutputResponse - the response from the model's run_input implementation.
    """

    if not runner_item.HasField('post_model_outputs_request'):
      raise Exception("Unexpected work item type: {}".format(runner_item))
    request = runner_item.post_model_outputs_request
    ensure_urls_downloaded(request)

    resp = self.predict_wrapper(request)
    successes = [o.status.code == status_code_pb2.SUCCESS for o in resp.outputs]
    if all(successes):
      status = status_pb2.Status(
          code=status_code_pb2.SUCCESS,
          description="Success",
      )
    elif any(successes):
      status = status_pb2.Status(
          code=status_code_pb2.MIXED_STATUS,
          description="Mixed Status",
      )
    else:
      status = status_pb2.Status(
          code=status_code_pb2.FAILURE,
          description="Failed",
      )

    resp.status.CopyFrom(status)
    return service_pb2.RunnerItemOutput(multi_output_response=resp)

  def runner_item_generate(
      self, runner_item: service_pb2.RunnerItem) -> Iterator[service_pb2.RunnerItemOutput]:
    # Call the generate() method the underlying model implements.

    if not runner_item.HasField('post_model_outputs_request'):
      raise Exception("Unexpected work item type: {}".format(runner_item))
    request = runner_item.post_model_outputs_request
    ensure_urls_downloaded(request)

    for resp in self.generate_wrapper(request):
      successes = []
      for output in resp.outputs:
        if not output.HasField('status') or not output.status.code:
          raise Exception("Output must have a status code, please check the model implementation.")
        successes.append(output.status.code == status_code_pb2.SUCCESS)
      if all(successes):
        status = status_pb2.Status(
            code=status_code_pb2.SUCCESS,
            description="Success",
        )
      elif any(successes):
        status = status_pb2.Status(
            code=status_code_pb2.MIXED_STATUS,
            description="Mixed Status",
        )
      else:
        status = status_pb2.Status(
            code=status_code_pb2.FAILURE,
            description="Failed",
        )
      resp.status.CopyFrom(status)

      yield service_pb2.RunnerItemOutput(multi_output_response=resp)

  def runner_item_stream(self, runner_item_iterator: Iterator[service_pb2.RunnerItem]
                        ) -> Iterator[service_pb2.RunnerItemOutput]:
    # Call the generate() method the underlying model implements.
    for resp in self.stream_wrapper(pmo_iterator(runner_item_iterator)):
      successes = []
      for output in resp.outputs:
        if not output.HasField('status') or not output.status.code:
          raise Exception("Output must have a status code, please check the model implementation.")
        successes.append(output.status.code == status_code_pb2.SUCCESS)
      if all(successes):
        status = status_pb2.Status(
            code=status_code_pb2.SUCCESS,
            description="Success",
        )
      elif any(successes):
        status = status_pb2.Status(
            code=status_code_pb2.MIXED_STATUS,
            description="Mixed Status",
        )
      else:
        status = status_pb2.Status(
            code=status_code_pb2.FAILURE,
            description="Failed",
        )
      resp.status.CopyFrom(status)

      yield service_pb2.RunnerItemOutput(multi_output_response=resp)


def pmo_iterator(runner_item_iterator):
  for runner_item in runner_item_iterator:
    if not runner_item.HasField('post_model_outputs_request'):
      raise Exception("Unexpected work item type: {}".format(runner_item))
    ensure_urls_downloaded(runner_item.post_model_outputs_request)
    yield runner_item.post_model_outputs_request



from typing import Any, Dict, List

from clarifai_grpc.grpc.api import resources_pb2
from clarifai_grpc.grpc.api.service_pb2 import MultiModelTypeResponse
from google.protobuf.json_format import MessageToDict
from google.protobuf.struct_pb2 import Struct


def response_to_templates(response: MultiModelTypeResponse, model_type_id: str) -> List[str]:
  """Converts the response from the API to a list of templates for the given model type id."""
  dict_response = MessageToDict(response)
  templates = []
  for model_type in dict_response['modelTypes']:
    if model_type['id'] == model_type_id:
      for modeltypefield in model_type['modelTypeFields']:
        if modeltypefield['path'].split('.')[-1] == "template":
          templates = [template['id'] for template in modeltypefield['modelTypeEnumOptions']]
  return templates


def response_to_model_params(response: MultiModelTypeResponse,
                             model_type_id: str,
                             template: str = None) -> Dict[str, Any]:
  """Converts the response from the API to a dictionary of model params for the given model type id."""
  dict_response = MessageToDict(response)
  params = {}
  if model_type_id != "clusterer":
    params["dataset_id"] = ""
    params["dataset_version_id"] = ""
  if model_type_id not in ["clusterer", "text-to-text"]:
    params["concepts"] = []
  params["train_params"] = dict()

  for model_type in dict_response['modelTypes']:
    if model_type['id'] == model_type_id:
      #iterate through the model type fields
      for modeltypefield in model_type['modelTypeFields']:
        _path = modeltypefield['path'].split('.')
        #removing the fields which are not required
        if (_path[0] in ["'eval_info'"]) or (_path[1] in ["dataset", "data"]) or (_path[-1] in [
            "dataset_id", "dataset_version_id"
        ]) or ("internalOnly" in modeltypefield.keys()):
          continue
        #checking the template model type fields
        if _path[-1] != "template":
          if _path[0] == 'train_info':
            try:
              params["train_params"][_path[-1]] = modeltypefield['defaultValue']
            except Exception:
              params["train_params"][_path[-1]] = None
          if _path[0] == 'output_info':
            params["inference_params"] = dict()
            try:
              params["inference_params"][_path[-1]] = modeltypefield['defaultValue']
            except Exception:
              params["inference_params"][_path[-1]] = None
          if _path[0] == 'output_info' and _path[1] == 'output_config':
            params["concepts_mutually_exclusive"] = (modeltypefield['defaultValue'])
        else:
          if 'modelTypeEnumOptions' in modeltypefield.keys():
            #check given template is valid
            all_templates = [template['id'] for template in modeltypefield['modelTypeEnumOptions']]
            if template not in all_templates:
              raise ValueError(f"Invalid template {template} for model type {model_type_id}. "
                               f"Valid templates are {all_templates}")
            for modeltypeenum in modeltypefield['modelTypeEnumOptions']:
              #finding the given template
              if modeltypeenum['id'] == template:
                params['train_params']["template"] = modeltypeenum['id']
                #iterate through the template fields
                for modeltypeenumfield in modeltypeenum['modelTypeFields']:
                  if "internalOnly" in modeltypeenumfield.keys():
                    continue
                  try:
                    params["train_params"][modeltypeenumfield['path'].split('.')[
                        -1]] = modeltypeenumfield['defaultValue']
                  except Exception:
                    params["train_params"][modeltypeenumfield['path'].split('.')[-1]] = None
  #custom config
  if "custom_config" in params['train_params'].keys():
    # Write the content to the file
    file_path = params['train_params']['template'] + ".py"
    with open(file_path, "w") as script_file:
      script_file.write(params['train_params']['custom_config'])
    params['train_params']['custom_config'] = file_path

  return params


def params_parser(params_dict: dict, concepts: List = None) -> Dict[str, Any]:
  """Converts the params dictionary to a dictionary of model specific params for the given model"""
  #dict parser
  train_dict = {}
  train_dict["train_info"] = dict()
  train_dict["output_info"] = dict()

  train_dict["train_info"]['params'] = Struct()
  if 'custom_config' in params_dict['train_params'].keys():
    # Open and read the Python file
    with open(params_dict['train_params']['custom_config'], 'r') as python_file:
      custom_config = python_file.read()
    params_dict['train_params']['custom_config'] = custom_config

  train_dict["train_info"]['params'].update(params_dict["train_params"])
  if 'dataset_id' in params_dict.keys():
    train_dict["train_info"]['params']['dataset_id'] = params_dict['dataset_id']
    train_dict["train_info"]['params']['dataset_version_id'] = params_dict['dataset_version_id']
  train_dict['train_info'] = resources_pb2.TrainInfo(**train_dict['train_info'])

  if 'concepts' in params_dict.keys():
    assert set(params_dict["concepts"]).issubset(
        concepts), "Invalid concept IDs. Available concepts in the app are {}".format(concepts)
    train_dict["output_info"]['data'] = resources_pb2.Data(
        concepts=[resources_pb2.Concept(id=concept_id) for concept_id in params_dict["concepts"]])
  if 'inference_params' in params_dict.keys():
    train_dict["output_info"]['params'] = Struct()
    train_dict['output_info']['params'].update(params_dict["inference_params"])
  if 'concepts_mutually_exclusive' in params_dict.keys():
    train_dict['output_info']['output_config'] = resources_pb2.OutputConfig(
        concepts_mutually_exclusive=params_dict['concepts_mutually_exclusive'])
  train_dict['output_info'] = resources_pb2.OutputInfo(**train_dict['output_info'])

  return train_dict


def response_to_param_info(response: MultiModelTypeResponse,
                           model_type_id: str,
                           param: str,
                           template: str = None) -> Dict[str, Any]:
  """Converts the response from the API to a dictionary of model param info for the given model type id."""
  dict_response = MessageToDict(response)
  for model_type in dict_response['modelTypes']:
    if model_type['id'] == model_type_id:
      #iterate through the model type fields
      for modeltypefield in model_type['modelTypeFields']:
        if modeltypefield['path'].split('.')[-1] == param:
          if param == 'template':
            del modeltypefield['placeholder']
            del modeltypefield['modelTypeEnumOptions']
            return modeltypefield
          modeltypefield['param'] = modeltypefield.pop('path').split('.')[-1]
          del modeltypefield['placeholder']
          return modeltypefield
        #checking the template model type fields
        if modeltypefield['path'].split('.')[-1] == "template":
          for modeltypeenum in modeltypefield['modelTypeEnumOptions']:
            if modeltypeenum['id'] == template:
              #iterate through the template fields
              for modeltypeenumfield in modeltypeenum['modelTypeFields']:
                if modeltypeenumfield['path'].split('.')[-1] == param:
                  modeltypeenumfield['param'] = modeltypeenumfield.pop('path').split('.')[-1]
                  del modeltypeenumfield['placeholder']
                  return modeltypeenumfield


def find_and_replace_key(nested_dict: Dict, target_key: str, replacement_value: Any) -> None:
  """Finds and replaces the target key with the replacement value in the nested dictionary."""
  for key, value in nested_dict.items():
    if key == target_key:
      nested_dict[key] = replacement_value
    elif isinstance(value, dict):
      find_and_replace_key(value, target_key, replacement_value)
